from __future__ import annotations
from time import time
import pygame
import random
from pygame import mask
from pygame.mixer import Sound
from pygame.math import Vector2
from pygame import mouse as muse
from math import atan2, pi
from typing import Union

class Actor:
    def __init__(self, window: pygame.Surface, image: Union[pygame.Surface, str], x: float = 0, y: float = 0, gravity: float = 0, width: int = None, height: int = None, camera: Camera = None):
        self.window = window
        if isinstance(image, pygame.Surface):
            self.image = image
        else:
            self.image = pygame.image.load(image)
        self.gravity = gravity
        if width is not None and height is not None:
            self.image = pygame.transform.scale(self.image, (width, height))
        self.width = self.image.get_width()
        self.height = self.image.get_height()
        self.floor = window.get_height()
        self.mask = mask.from_surface(self.image)
        self.camera = camera
        self.dead = False
        self.x = x
        self.y = y

    def move(self, x: float, y: float):
        if self.dead:
            return
        self.x += x
        self.y += y
        return (self.x, self.y)

    def setImage(self, image: Union[pygame.Surface, str], width: int = None, height: int = None, flip: bool = False):
        if isinstance(image, pygame.Surface):
            self.image = image
        else:
            self.image = pygame.image.load(image)
        if width is not None and height is not None:
            self.image = pygame.transform.scale(self.image, (width, height))
        if flip:
            self.image = pygame.transform.flip(self.image, True, False)
        self.width = self.image.get_width()
        self.height = self.image.get_height()
        self.mask = mask.from_surface(self.image)

    def fall(self):
        if self.dead:
            return
        self.y += self.gravity
        self.y = min(self.y, self.floor - self.height)

    def overlap(self, actor: Actor):
        if self.dead:
            return False
        return self.mask.overlap(actor.mask, (round(actor.x) - round(self.x), round(actor.y) - round(self.y))) is not None

    def draw(self):
        if not self.dead:
            if self.camera is None:
                self.window.blit(self.image, (self.x, self.y))
            else:
                self.window.blit(self.image, (self.x - self.camera.x, self.y - self.camera.y))


class Bullet(Actor):
    def __init__(self, window: pygame.Surface, velocity: tuple[float, float] = (0, 0), targets: list[Actor] = [], x: float = 0, y: float = 0, camera: Camera = None):
        bullet = pygame.image.load('bullet.png')
        bullet = pygame.transform.rotate(bullet, -atan2(velocity[1], velocity[0]) * 180 / pi)
        Actor.__init__(self, window, bullet, x, y, 0, 40, 40, camera)
        self.velocity = velocity
        self.targets = targets

    def update(self):
        self.move(self.velocity[0], self.velocity[1])
        for i in self.targets:
            if self.overlap(i) and not i.dead:
                i.dead = True
                Sound('hit.wav').play()
        self.draw()

class Bunny(Actor):
    frames = ['bunnystanding.png', 'bunnyrun.png']
    def __init__(self, window: pygame.Surface, x: float = 0, y: float = 0, camera: Camera = None):
        Actor.__init__(self, window, 'bunnystanding.png', x, y, 1.5, 80, 80, camera)
        feet = pygame.Surface((self.width, self.height))
        feet.blit(self.image, (0, 0))
        self.feet = mask.Mask((self.width, self.height), fill=False)
        for x in range(10, self.width - 19):
            for y in range(75, self.height):
                self.feet.set_at((x, y), 1)
        self.velX = 0
        self.velY = 0
        self.flip = True
        self.sneak = False
        self.frame = 0
        self.framesElapsed = 0
        self.timeSinceLastShoot = time()
    
    def foverlap(self, actor: Actor):
        if self.dead:
            return False
        return self.feet.overlap(actor.mask, (round(actor.x) - round(self.x), round(actor.y) - round(self.y))) is not None

    def events(self, event):
        self.sneak = False
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_a:
                self.flip = False
                self.velX -= 1.75
            elif event.key == pygame.K_d:
                self.flip = True
                self.velX += 1.75
            elif event.key == pygame.K_SPACE and round(self.y) == self.floor - self.height:
                self.velY = -5
                Sound('jump.wav').play()
            elif event.key == pygame.K_LSHIFT or event.key == pygame.K_RSHIFT:
                self.sneak = True
        if event.type == pygame.KEYUP:
            if event.key == pygame.K_a:
                self.velX += 1.75
            elif event.key == pygame.K_d:
                self.velX -= 1.75

    def checkPlatforms(self, platforms: list[Platform]):
        for i in platforms:
            if self.foverlap(i) and i.y >= self.y:
                self.floor = i.y + 1
                return
        self.floor = 1000000

    def shoot(self, mouse: Mouse, eagles: list[Eagle]):
        if self.timeSinceLastShoot + 5 > time() or self.dead:
            return None
        self.timeSinceLastShoot = time()
        velocity = Vector2(mouse.x - self.x, mouse.y - self.y).normalize() * 6
        Sound('shoot.wav').play()
        return Bullet(self.window, (velocity.x, velocity.y), eagles, self.x, self.y, self.camera)

    def update(self):
        if self.dead:
            return
        if not self.y + 80 == self.floor:
            self.setImage('bunnyjump.png', 80, 80, self.flip)
        elif self.sneak:
            self.setImage('bunnyduck.png', 80, 80, self.flip)
        elif self.velX == 0:
            self.setImage('bunnystanding.png', 80, 80, self.flip)
        else:
            self.framesElapsed += 1
            if self.framesElapsed % 10 == 0:
                self.frame += 1
                if self.frame > 1:
                    self.frame = 0
            self.setImage(self.frames[self.frame], 80, 80, flip=self.flip)
        if self.velY < 0:
            self.velY += 0.1
        self.move((0 if self.sneak else self.velX), self.velY)
        if self.x < 0:
            self.x = 0
        self.camera.x = self.x - 360
        self.camera.y = self.y - 260
        if self.camera.x < 0:
            self.camera.x = 0
        self.fall()
        self.draw()

class Eagle(Actor):
    frames = ['eagle1.png', 'eagle2.png', 'eagle3.png', 'eagle2.png']

    def __init__(self, window: pygame.Surface, bunny: Bunny, x: float = 0, y: float = 0, camera: Camera = None):
        Actor.__init__(self, window, 'eagle1.png', x, y, 0, 80, 80, camera)
        self.bunny = bunny
        self.frame = 0
        self.framesElapsed = 0
        self.timeSinceLastShoot = time()
        self.flip = True
    
    def nextFrame(self):
        self.frame += 1
        if self.frame > 3:
            self.frame = 0
    
    def follow(self):
        if self.bunny.dead:
            return
        direction = Vector2(self.bunny.x - self.x, self.bunny.y - self.y).normalize() * (1.71 if self.bunny.x < self.x + 360 else 2)
        self.x += direction.x
        self.y += direction.y
        if direction.x < 0:
            self.flip = False
        else:
            self.flip = True

    def update(self, bunny: Bunny = None):
        if self.dead:
            return
        if bunny is not None:
            self.bunny = bunny
        if self.bunny.overlap(self):
            self.bunny.dead = True
        self.framesElapsed += 1
        if self.framesElapsed % 10 == 0:
            self.nextFrame()
        self.setImage(self.frames[self.frame], 80, 80, self.flip)
        self.follow()
        self.draw()


class Platform(Actor):
    def __init__(self, window: pygame.Surface, x: float = 0, y: float = 0, width: int = 10, camera: Camera = None):
        Actor.__init__(self, window, 'platform.png', x, y, 0, width, 2, camera)

class SpawnGround(Platform):
    def __init__(self, window: pygame.Surface, x: float = 0, y: float = 0, camera: Camera = None):
        Actor.__init__(self, window, 'platform.png', x, y, 0, 200, 800, camera)

class PlatformGenerator:
    def __init__(self, window: pygame.Surface, x: int, y: int, widths: list[int], platforms: int, camera: Camera = None):
        self.window = window
        self.startX = x
        self.startY = y
        self.widths = widths
        self.amount = platforms
        self.camera = camera

    def generate(self):
        platforms = [Platform(self.window, self.startX, self.startY, random.choice(self.widths), self.camera)]
        for i in range(1, self.amount):
            position = (platforms[i - 1].x + random.randint(110, 160), platforms[i - 1].y + random.randint(-30, -10))
            width = random.choice(self.widths)
            platforms.append(Platform(self.window, position[0], position[1], width, self.camera))
        return platforms

class Mouse:
    def __init__(self, camera: Camera):
        self.camera = camera
        self.update()
    
    def update(self):
        self.x = muse.get_pos()[0] + self.camera.x
        self.y = muse.get_pos()[1] + self.camera.y

class Camera:
    def __init__(self, x: float = 0, y: float = 0):
        self.x = x
        self.y = y
    
    def move(self, x: float, y: float):
        self.x += x
        self.y += y
